@@ -9,21 +9,29 @@ class urls :
         self.base_trade_url = 'https://tradeapi.webullbroker.com/api/trade'
         self.base_user_url = 'https://userapi.webull.com/api'
         self.base_userbroker_url = 'https://userapi.webullbroker.com/api'
+        self.base_ustrade_url = 'https://ustrade.webullfinance.com/api'
+        self.base_paperfintech_url = 'https://act.webullfintech.com/webull-paper-center/api'
+        self.base_fintech_gw_url = 'https://quotes-gw.webullfintech.com/api'
+        self.base_userfintech_url = 'https://userapi.webullfintech.com/api'
+
 
     def account(self, account_id):
         return f'{self.base_trade_url}/v2/home/{account_id}'
 
     def account_id(self):
-        return f'{self.base_trade_url}/account/getSecAccountList/v4'
+        return f'{self.base_trade_url}/account/getSecAccountList/v5'
 
-    def active_gainers_losers(self, direction):
-        if direction == 'gainer':
-            url = 'advanced'
-        elif direction == 'loser':
-            url = 'declined'
-        else:
-            url = 'active'
-        return f'{self.base_securities_url}/securities/market/v5/card/stockActivityPc.{url}/list'
+    def account_activities(self, account_id):
+        return f'{self.base_ustrade_url}/trade/v2/funds/{account_id}/activities'
+
+    def active_gainers_losers(self, direction, region_code, rank_type, num) :
+          if direction == 'gainer' :
+              url = 'topGainers'
+          elif direction == 'loser' :
+              url = 'dropGainers'
+          else :
+              url = 'topActive'
+          return f'{self.base_fintech_gw_url}/wlas/ranking/{url}?regionId={region_code}&rankType={rank_type}&pageIndex=1&pageSize={num}'
 
     def add_alert(self):
         return f'{self.base_userbroker_url}/user/warning/v2/manage/overlap'
@@ -31,15 +39,30 @@ class urls :
     def analysis(self, stock):
         return f'{self.base_securities_url}/securities/ticker/v5/analysis/{stock}'
 
+    def analysis_shortinterest(self, stock):
+        return f'{self.base_securities_url}/securities/stock/{stock}/shortInterest'
+
+    def analysis_institutional_holding(self, stock):
+        return f'{self.base_securities_url}/securities/stock/v5/{stock}/institutionalHolding'
+
+    def analysis_etf_holding(self, stock, has_num, page_size):
+        return f'{self.base_securities_url}/securities/stock/v5/{stock}/belongEtf?hasNum={has_num}&pageSize={page_size}'
+
+    def analysis_capital_flow(self, stock, show_hist):
+        return f'{self.base_securities_url}/wlas/capitalflow/ticker?tickerId={stock}&showHis={show_hist}'
+
     def bars(self, stock):
         return f'{self.base_quote_url}/quote/tickerChartDatas/v5/{stock}'
 
     def cancel_order(self, account_id):
         return f'{self.base_trade_url}/order/{account_id}/cancelStockOrder/'
 
-    def cancel_otoco_orders(self, account_id):
+    def modify_otoco_orders(self, account_id):
         return f'{self.base_trade_url}/v2/corder/stock/modify/{account_id}'
 
+    def cancel_otoco_orders(self, account_id, combo_id):
+        return f'{self.base_trade_url}/v2/corder/stock/cancel/{account_id}/{combo_id}'
+
     def check_otoco_orders(self, account_id):
         return f'{self.base_trade_url}/v2/corder/stock/check/{account_id}'
 
@@ -56,10 +79,19 @@ class urls :
         return f'{self.base_userbroker_url}/user/warning/v2/query/tickers'
 
     def login(self):
-        return f'{self.base_user_url}/passport/login/v3/account'
+        return f'{self.base_user_url}/passport/login/v5/account'
 
-    def get_mfa(self, account, account_type, device_id, code_type, region_code):
-        return f'{self.base_user_url}/passport/verificationCode/sendCode?account={account}&accountType={account_type}&deviceId={device_id}&codeType={code_type}&regionCode={region_code}'
+    def get_mfa(self) :
+        return f'{self.base_userbroker_url}/passport/v2/verificationCode/send'
+
+    def check_mfa(self) :
+        return f'{self.base_userfintech_url}/passport/v2/verificationCode/checkCode'
+
+    def get_security(self, username, account_type, region_code, event) :
+        return f'{self.base_userfintech_url}/user/risk/getPrivacyQuestion?account={username}&accountType={account_type}&regionId={region_code}&event={event}'
+
+    def check_security(self) :
+        return f'{self.base_userfintech_url}/user/risk/checkAnswer'
 
     def logout(self):
         return f'{self.base_userbroker_url}/passport/login/logout'
@@ -83,7 +115,7 @@ class urls :
         return f'{self.base_paper_url}/paper/1/acc/{paper_account_id}/order?&startTime=1970-0-1&dateType=ORDER&pageSize={page_size}&status='
 
     def paper_account(self, paper_account_id):
-        return f'{self.base_paper_url}/paper/1/acc/{paper_account_id}'
+        return f'{self.base_paperfintech_url}/paper/1/acc/{paper_account_id}'
 
     def paper_account_id(self):
         return f'{self.base_paper_url}/myaccounts/true'
@@ -98,7 +130,7 @@ class urls :
         return f'{self.base_paper_url}/paper/1/acc/{paper_account_id}/orderop/place/{stock}'
 
     def place_option_orders(self, account_id):
-        return f'{self.base_trade_url}/v2/option/placeOrder/{account_id}'
+        return f'{self.base_ustrade_url}/trade/v2/option/placeOrder/{account_id}'
 
     def place_orders(self, account_id):
         return f'{self.base_trade_url}/order/{account_id}/placeStockOrder'
@@ -124,8 +156,8 @@ class urls :
     def replace_option_orders(self, account_id):
         return f'{self.base_trade_url}/v2/option/replaceOrder/{account_id}'
 
-    def stock_id(self, stock):
-        return f'{self.base_options_gw_url}/search/pc/tickers?keyword={stock}&pageIndex=1&pageSize=1'
+    def stock_id(self, stock, region_code):
+        return f'{self.base_options_gw_url}/search/pc/tickers?keyword={stock}&pageIndex=1&pageSize=20&regionId={region_code}'
 
     def trade_token(self):
         return f'{self.base_trade_url}/login'
@@ -141,3 +173,6 @@ class urls :
 
     def social_home(self, topic, num=100):
         return f'{self.base_user_url}/social/feed/topic/{topic}/home?size={num}'
+
+    def portfolio_lists(self):
+        return f'{self.base_options_gw_url}/personal/portfolio/v2/check'
