@@ -8,6 +8,7 @@ import pickle
 import requests
 import time
 import uuid
+import urllib.parse
 
 from datetime import datetime, timedelta
 from email_validator import validate_email, EmailNotValidError
@@ -25,6 +26,10 @@ class webull:
             'Accept': '*/*',
             'Accept-Encoding': 'gzip, deflate',
             'Content-Type': 'application/json',
+            'platform': 'web',
+            'ver': '3.21.14',
+            'User-Agent': '*',
+            'did': self._get_did(),
         }
         self._access_token = ''
         self._account_id = ''
@@ -32,19 +37,25 @@ class webull:
         self._token_expire = ''
         self._trade_token = ''
         self._uuid = ''
+        self._region_code = 6
         self._did = self._get_did()
         self._urls = endpoints.urls()
 
-    def _get_did(self):
+    def _get_did(self, path=''):
         '''
         Makes a unique device id from a random uuid (uuid.uuid4).
         if the pickle file doesn't exist, this func will generate a random 32 character hex string
         uuid and save it in a pickle file for future use. if the file already exists it will
         load the pickle file to reuse the did. Having a unique did appears to be very important
         for the MQTT web socket protocol
+
+        path: path to did.bin. For example _get_did('cache') will search for cache/did.bin instead.
+
         :return: hex string of a 32 digit uuid
         '''
         filename = 'did.bin'
+        if path:
+            filename = os.path.join(path, filename)
         if os.path.exists(filename):
             did = pickle.load(open(filename,'rb'))
         else:
@@ -67,7 +78,7 @@ class webull:
         return headers
 
 
-    def login(self, username='', password='', device_name='', mfa=''):
+    def login(self, username='', password='', device_name='', mfa='', question_id='', question_answer=''):
         '''
         Login with email or phone number
 
@@ -93,17 +104,21 @@ class webull:
             device_name = 'default_string'
 
         data = {
+            'accessQuestions': '[{"questionId":"' + str(question_id) + '", "answer":"' + str(question_answer) + '"}]',
             'account': username,
             'accountType': accountType,
             'deviceId': self._did,
             'deviceName': device_name,
             'grade': 1,
             'pwd': md5_hash.hexdigest(),
-            'regionId': 1
+            'regionId': self._region_code
         }
 
         if mfa != '' :
-            data['extInfo'] = {'verificationCode': mfa}
+            data['extInfo'] = {
+                'codeAccountType': accountType,
+                'verificationCode': mfa
+            }
             headers = self.build_req_headers()
         else :
             headers = self._headers
@@ -117,22 +132,72 @@ class webull:
             self._account_id = self.get_account_id()
         return result
 
-    def api_login(self, access_token='', refresh_token='', token_expire='', uuid='', mfa=''):
-        self._access_token = access_token
-        self._refresh_token = refresh_token
-        self._token_expire = token_expire
-        self._uuid = uuid
-        self._account_id = self.get_account_id()
-
-    def get_mfa(self, username):
+    def get_mfa(self, username='') :
         try:
           validate_email(username)
           accountType = 2 # email
         except EmailNotValidError as _e:
           accountType = 1 # phone
 
-        response = requests.get(self._urls.get_mfa(username, str(accountType), str(self._did), str(5), str(1)), headers=self._headers)
+        data = {'account': str(username),
+                'accountType': str(accountType),
+                'codeType': int(5)}
 
+        response = requests.post(self._urls.get_mfa(), json=data, headers=self._headers)
+        # data = response.json()
+        
+        if response.status_code == 200 :
+            return True
+        else :
+            return False
+
+    def check_mfa(self, username='', mfa='') :
+        try:
+          validate_email(username)
+          accountType = 2 # email
+        except EmailNotValidError as _e:
+          accountType = 1 # phone
+
+        data = {'account': str(username),
+                'accountType': str(accountType),
+                'code': str(mfa),
+                'codeType': int(5)}
+
+        response = requests.post(self._urls.check_mfa(), json=data, headers=self._headers)
+        data = response.json()
+
+        return data
+
+    def get_security(self, username='') :
+        try:
+          validate_email(username)
+          accountType = 2 # email
+        except EmailNotValidError as _e:
+          accountType = 1 # phone
+
+        username = urllib.parse.quote(username)
+
+        response = requests.get(self._urls.get_security(username, accountType, self._region_code, 'PRODUCT_LOGIN'), headers=self._headers)
+        data = response.json()
+
+        return data
+
+    def check_security(self, username='', question_id='', question_answer='') :
+        try:
+          validate_email(username)
+          accountType = 2 # email
+        except EmailNotValidError as _e:
+          accountType = 1 # phone
+
+        data = {'account': str(username),
+                'accountType': str(accountType),
+                'answerList': [{'questionId': str(question_id), 'answer': str(question_answer)}],
+                'event': 'PRODUCT_LOGIN'}
+
+        response = requests.post(self._urls.check_security(), json=data, headers=self._headers)
+        data = response.json()
+
+        return data
 
     def login_prompt(self):
         '''
@@ -152,6 +217,13 @@ class webull:
         response = requests.get(self._urls.logout(), headers=headers)
         return response.status_code
 
+    def api_login(self, access_token='', refresh_token='', token_expire='', uuid='', mfa=''):
+        self._access_token = access_token
+        self._refresh_token = refresh_token
+        self._token_expire = token_expire
+        self._uuid = uuid
+        self._account_id = self.get_account_id()
+
     def refresh_login(self):
         '''
         Refresh login token
@@ -167,7 +239,6 @@ class webull:
             self._token_expire = result['tokenExpireTime']
         return result
 
-
     def get_detail(self):
         '''
         get some contact details of your account name, email/phone, region, avatar...etc
@@ -179,7 +250,6 @@ class webull:
 
         return result
 
-
     def get_account_id(self):
         '''
         get account id
@@ -191,11 +261,11 @@ class webull:
         result = response.json()
         if result['success']:
             id = str(result['data'][0]['secAccountId'])
+            self._account_id = id
             return id
         else:
             return None
 
-
     def get_account(self):
         '''
         get important details of account, positions, portfolio stance...etc
@@ -205,7 +275,6 @@ class webull:
         result = response.json()
         return result
 
-
     def get_positions(self):
         '''
         output standing positions of stocks
@@ -213,7 +282,6 @@ class webull:
         data = self.get_account()
         return data['positions']
 
-
     def get_portfolio(self):
         '''
         output numbers of portfolio
@@ -224,15 +292,23 @@ class webull:
             output[item['key']] = item['value']
         return output
 
+    def get_activities(self, index=1, size=500) :
+        '''
+        Activities including transfers, trades and dividends
+        '''
+        headers = self.build_req_headers(include_trade_token=True, include_time=True)
+        data = {'pageIndex': index,
+                'pageSize': size}
+        response = requests.post(self._urls.account_activities(self._account_id), json=data, headers=headers)
+        return response.json()
 
-    def get_current_orders(self):
+    def get_current_orders(self) :
         '''
         Get open/standing orders
         '''
         data = self.get_account()
         return data['openOrders']
 
-
     def get_history_orders(self, status='Cancelled', count=20):
         '''
         Historical orders, can be cancelled or filled
@@ -242,7 +318,6 @@ class webull:
         response = requests.get(self._urls.orders(self._account_id, count) + str(status), headers=headers)
         return response.json()
 
-
     def get_trade_token(self, password=''):
         '''
         Trading related
@@ -263,35 +338,65 @@ class webull:
         else:
             return False
 
-
     def get_ticker(self, stock=''):
         '''
         Lookup ticker_id
+        Ticker issue, will attempt to find an exact match, if none is found, match the first one
         '''
         headers = self.build_req_headers()
         ticker_id = 0
         if stock and isinstance(stock, str):
-            response = requests.get(self._urls.stock_id(stock), headers=headers)
+            response = requests.get(self._urls.stock_id(stock, self._region_code), headers=headers)
             result = response.json()
-            if result.get('data'):
-                for item in result['data']: # implies multiple tickers, but only assigns last one?
-                    ticker_id = item['tickerId']
+            if result.get('data') :
+                for item in result['data'] : # implies multiple tickers, but only assigns last one?
+                    if 'symbol' in item and item['symbol'] == stock :
+                        ticker_id = item['tickerId']
+                        break
+                    elif 'disSymbol' in item and item['disSymbol'] == stock :
+                        ticker_id = item['tickerId']
+                        break
+                if ticker_id == 0 :
+                    ticker_id = result['data'][0]['tickerId']
             else:
                 raise ValueError('TickerId could not be found for stock {}'.format(stock))
         else:
             raise ValueError('Stock symbol is required')
         return ticker_id
 
+    '''
+    Actions related to stock
+    '''
+    def get_quote(self, stock=None, tId=None):
+        '''
+        get price quote
+        tId: ticker ID str
+        '''
+        headers = self.build_req_headers()
+        if not stock and not tId:
+            raise ValueError('Must provide a stock symbol or a stock id')
 
-    def place_order(self, stock=None, tId=None, price=0, action='BUY',
-            orderType='LMT', enforce='GTC', qty=0, outsideRegularTradingHour=True):
+        if stock:
+            try:
+                tId = str(self.get_ticker(stock))
+            except ValueError as _e:
+                raise ValueError("Could not find ticker for stock {}".format(stock))
+        response = requests.get(self._urls.quotes(tId), headers=headers)
+        result = response.json()
+        return result
+
+    def place_order(self, stock=None, tId=None, price=0, action='BUY', orderType='LMT', enforce='GTC', quant=0, outsideRegularTradingHour=True, stpPrice=None, trial_value=0, trial_type='DOLLAR'):
         '''
         Place an order
 
+        price: float (LMT / STP LMT Only)
         action: BUY / SELL
-        ordertype : LMT / MKT / STP / STP LMT
+        ordertype : LMT / MKT / STP / STP LMT / STP TRAIL
         timeinforce:  GTC / DAY / IOC
         outsideRegularTradingHour: True / False
+        stpPrice: float (STP / STP LMT Only)
+        trial_value: float (STP TRIAL Only)
+        trial_type: DOLLAR / PERCENTAGE (STP TRIAL Only)
         '''
         if not tId is None:
             pass
@@ -303,29 +408,37 @@ class webull:
         headers = self.build_req_headers(include_trade_token=True, include_time=True)
         data = {
             'action': action,
-            'lmtPrice': float(price),
             'orderType': orderType,
             'outsideRegularTradingHour': outsideRegularTradingHour,
-            'quantity': int(qty),
+            'quantity': int(quant),
             'serialId': str(uuid.uuid4()),
             'tickerId': tId,
             'timeInForce': enforce
         }
 
         # Market orders do not support extended hours trading.
-        if orderType == 'MKT':
+        if orderType == 'MKT' :
             data['outsideRegularTradingHour'] = False
+        elif orderType == 'LMT':
+            data['lmtPrice'] = float(price)
+        elif orderType == 'STP' :
+            data['auxPrice'] = float(stpPrice)
+        elif orderType == 'STP LMT' :
+            data['lmtPrice'] = float(price)
+            data['auxPrice'] = float(stpPrice)
+        elif orderType == 'STP TRAIL' :
+            data['trailingStopStep'] = float(trial_value)
+            data['trailingType'] = str(trial_type)
 
         response = requests.post(self._urls.place_orders(self._account_id), json=data, headers=headers)
         return response.json()
 
-
     def modify_order(self, order=None, price=0, action=None, orderType=None, enforce=None, quant=0, outsideRegularTradingHour=None):
         '''
         Modify an order
 
         action: BUY / SELL
-        ordertype : LMT / MKT / STP / STP LMT
+        ordertype : LMT / MKT / STP / STP LMT / STP TRAIL
         timeinforce:  GTC / DAY / IOC
         outsideRegularTradingHour: True / False
         '''
@@ -360,9 +473,17 @@ class webull:
 
         return response.json()
 
+    def cancel_order(self, order_id=''):
+        '''
+        Cancel an order
+        '''
+        headers = self.build_req_headers(include_trade_token=True, include_time=True)
+        data = {}
+        response = requests.post(self._urls.cancel_order(self._account_id) + str(order_id) + '/' + str(uuid.uuid4()), json=data, headers=headers)
+        result = response.json()
+        return result['success']
 
-    def place_otoco_order(self, stock='', price='', stop_loss_price='', limit_profit_price='',
-        time_in_force='DAY', quant=0):
+    def place_order_otoco(self, stock='', price='', stop_loss_price='', limit_profit_price='', time_in_force='DAY', quant=0) :
         '''
         OTOCO: One-triggers-a-one-cancels-the-others, aka Bracket Ordering
         Submit a buy order, its fill will trigger sell order placement. If one sell fills, it will cancel the other
@@ -386,77 +507,107 @@ class webull:
         response1 = requests.post(self._urls.check_otoco_orders(self._account_id), json=data1, headers=headers)
         result1 = response1.json()
 
-        if result1['forward']:
-            data2 = {
-                'newOrders': [
-                    {'orderType': 'LMT', 'timeInForce': time_in_force, 'quantity': int(quant),
-                     'outsideRegularTradingHour': False, 'action': 'BUY', 'tickerId': self.get_ticker(stock),
-                     'lmtPrice': float(price), 'comboType': 'MASTER', 'serialId': str(uuid.uuid4())},
-                    {'orderType': 'STP', 'timeInForce': time_in_force, 'quantity': int(quant),
-                     'outsideRegularTradingHour': False, 'action': 'SELL', 'tickerId': self.get_ticker(stock),
-                     'auxPrice': float(stop_loss_price), 'comboType': 'STOP_LOSS', 'serialId': str(uuid.uuid4())},
-                    {'orderType': 'LMT', 'timeInForce': time_in_force, 'quantity': int(quant),
-                     'outsideRegularTradingHour': False, 'action': 'SELL', 'tickerId': self.get_ticker(stock),
-                     'lmtPrice': float(limit_profit_price), 'comboType': 'STOP_PROFIT', 'serialId': str(uuid.uuid4())}],
-                    'serialId': str(uuid.uuid4())
-                }
+        if result1['forward'] :
+            data2 = {'newOrders': [
+                            {'orderType': 'LMT', 'timeInForce': time_in_force, 'quantity': int(quant),
+                             'outsideRegularTradingHour': False, 'action': 'BUY', 'tickerId': self.get_ticker(stock),
+                             'lmtPrice': float(price), 'comboType': 'MASTER', 'serialId': str(uuid.uuid4())},
+                            {'orderType': 'STP', 'timeInForce': time_in_force, 'quantity': int(quant),
+                             'outsideRegularTradingHour': False, 'action': 'SELL', 'tickerId': self.get_ticker(stock),
+                             'auxPrice': float(stop_loss_price), 'comboType': 'STOP_LOSS', 'serialId': str(uuid.uuid4())},
+                            {'orderType': 'LMT', 'timeInForce': time_in_force, 'quantity': int(quant),
+                             'outsideRegularTradingHour': False, 'action': 'SELL', 'tickerId': self.get_ticker(stock),
+                             'lmtPrice': float(limit_profit_price), 'comboType': 'STOP_PROFIT', 'serialId': str(uuid.uuid4())}],
+                            'serialId': str(uuid.uuid4())
+                    }
 
             response2 = requests.post(self._urls.place_otoco_orders(self._account_id), json=data2, headers=headers)
 
-            print('Resp 2: {}'.format(response2))
-            return True
+            # print('Resp 2: {}'.format(response2))
+            return response2.text
         else:
             print(result1['checkResultList'][0]['code'])
             print(result1['checkResultList'][0]['msg'])
             return False
 
-
-    def cancel_order(self, order_id=''):
+    def modify_order_otoco(self, order_id1='', order_id2='', order_id3='', stock='', price='', stop_loss_price='', limit_profit_price='', time_in_force='DAY', quant=0) :
         '''
-        Cancel an order
+        OTOCO: One-triggers-a-one-cancels-the-others, aka Bracket Ordering
+        Submit a buy order, its fill will trigger sell order placement. If one sell fills, it will cancel the other
+         sell
         '''
-        headers = self.build_req_headers(include_trade_token=True, include_time=True)
-        data = {}
-        response = requests.post(self._urls.cancel_order(self._account_id) + str(order_id) + '/' + str(uuid.uuid4()), json=data, headers=headers)
-        result = response.json()
-        return result['success']
+        headers = self.build_req_headers(include_trade_token=False, include_time=True)
 
+        data = {'modifyOrders': [
+                        {'orderType': 'LMT', 'timeInForce': time_in_force, 'quantity': int(quant), 'orderId': str(order_id1),
+                         'outsideRegularTradingHour': False, 'action': 'BUY', 'tickerId': self.get_ticker(stock),
+                         'lmtPrice': float(price), 'comboType': 'MASTER', 'serialId': str(uuid.uuid4())},
+                        {'orderType': 'STP', 'timeInForce': time_in_force, 'quantity': int(quant), 'orderId': str(order_id2),
+                         'outsideRegularTradingHour': False, 'action': 'SELL', 'tickerId': self.get_ticker(stock),
+                         'auxPrice': float(stop_loss_price), 'comboType': 'STOP_LOSS', 'serialId': str(uuid.uuid4())},
+                        {'orderType': 'LMT', 'timeInForce': time_in_force, 'quantity': int(quant), 'orderId': str(order_id3),
+                         'outsideRegularTradingHour': False, 'action': 'SELL', 'tickerId': self.get_ticker(stock),
+                         'lmtPrice': float(limit_profit_price), 'comboType': 'STOP_PROFIT', 'serialId': str(uuid.uuid4())}],
+                        'serialId': str(uuid.uuid4())
+                }
 
-    def cancel_all_orders(self):
-        '''
-        Cancels all open (aka 'working') orders
-        '''
-        open_orders = self.get_current_orders()
-        for order in open_orders:
-            self.cancel_order(order['orderId'])
+        response = requests.post(self._urls.modify_otoco_orders(self._account_id), json=data, headers=headers)
 
-    def cancel_otoco_order(self, order_id=''):
+        # print('Resp: {}'.format(response))
+        return response.text
+
+    def cancel_order_otoco(self, combo_id=''):
         '''
         Retract an otoco order. Cancelling the MASTER order_id cancels the sub orders.
         '''
         headers = self.build_req_headers(include_trade_token=True, include_time=True)
-        data = { 'serialId': str(uuid.uuid4()), 'cancelOrders': [str(order_id)]}
-        response = requests.post(self._urls.cancel_otoco_orders(self._account_id), json=data, headers=headers)
+        # data = { 'serialId': str(uuid.uuid4()), 'cancelOrders': [str(order_id)]}
+        data = {}
+        response = requests.post(self._urls.cancel_otoco_orders(self._account_id, combo_id), json=data, headers=headers)
         return response.json()
 
-    def get_quote(self, stock=None, tId=None):
+    '''
+    Actions related to cryptos
+    '''
+    def place_order_crypto(self, stock=None, tId=None, price=0, action='BUY', orderType='LMT', enforce='DAY', entrust_type='QTY', quant=0, outsideRegularTradingHour=False) :
         '''
-        get price quote
-        tId: ticker ID str
+        Place Crypto order
+        price: Limit order entry price
+        quant: dollar amount to buy/sell when entrust_type is CASH else the decimal or fractional amount of shares to buy
+        action: BUY / SELL
+        entrust_type: CASH / QTY
+        ordertype : LMT / MKT
+        timeinforce:  DAY
+        outsideRegularTradingHour: True / False
         '''
-        headers = self.build_req_headers()
-        if not stock and not tId:
+        if not tId is None:
+            pass
+        elif not stock is None:
+            tId = self.get_ticker(stock)
+        else:
             raise ValueError('Must provide a stock symbol or a stock id')
 
-        if stock:
-            try:
-                tId = str(self.get_ticker(stock))
-            except ValueError as _e:
-                raise ValueError("Could not find ticker for stock {}".format(stock))
-        response = requests.get(self._urls.quotes(tId), headers=headers)
-        result = response.json()
-        return result
+        headers = self.build_req_headers(include_trade_token=True, include_time=True)
+        data = {
+            'action': action,
+            'assetType': 'crypto',
+            'comboType': 'NORMAL',
+            'entrustType': entrust_type,
+            'lmtPrice': str(price),
+            'orderType': orderType,
+            'outsideRegularTradingHour': outsideRegularTradingHour,
+            'quantity': str(quant),
+            'serialId': str(uuid.uuid4()),
+            'tickerId': tId,
+            'timeInForce': enforce
+        }
 
+        response = requests.post(self._urls.place_orders(self._account_id), json=data, headers=headers)
+        return response.json()
+
+    '''
+    Actions related to options
+    '''
     def get_option_quote(self, stock=None, tId=None, optionId=None):
         '''
         get option quote
@@ -502,20 +653,27 @@ class webull:
                     expireDate = d['date']
                     break
 
-        params = {'count': count, 'includeWeekly': includeWeekly, 'direction': direction,
-                'expireDate': expireDate, 'unSymbol': stock, 'queryAll': queryAll}
-        return requests.get(self._urls.options(self.get_ticker(stock)), params=params, headers=headers).json()['data']
+        params = {'count': count,
+                  'includeWeekly': includeWeekly,
+                  'direction': direction,
+                  'expireDate': expireDate,
+                  'unSymbol': stock,
+                  'queryAll': queryAll}
+
+        data = requests.get(self._urls.options(self.get_ticker(stock)), params=params, headers=headers).json()
+
+        return data['data']
 
     def get_options_by_strike_and_expire_date(self, stock=None, expireDate=None, strike=None, direction='all'):
         '''
         get a list of options contracts by expire date and strike price
         strike: string
         '''
-        headers = self.build_req_headers()
-        opts = self.get_options(stock=stock, expireDate=expireDate, direction=direction, headers=headers)
+
+        opts = self.get_options(stock=stock, expireDate=expireDate, direction=direction)
         return [c for c in opts if c['strikePrice'] == strike]
 
-    def place_option_order(self, optionId=None, lmtPrice=None, stpPrice=None, action=None, orderType='LMT', enforce='DAY', quant=0):
+    def place_order_option(self, optionId=None, lmtPrice=None, stpPrice=None, action=None, orderType='LMT', enforce='DAY', quant=0):
         '''
         create buy / sell order
         stock: string
@@ -523,8 +681,8 @@ class webull:
         stpPrice: float
         action: string BUY / SELL
         optionId: string
-        orderType: LMT / STP / STP LMT
-        enforce: DAY
+        orderType: MKT / LMT / STP / STP LMT
+        enforce: GTC / DAY
         quant: int
         '''
         headers = self.build_req_headers(include_trade_token=True, include_time=True)
@@ -532,28 +690,28 @@ class webull:
             'orderType': orderType,
             'serialId': str(uuid.uuid4()),
             'timeInForce': enforce,
-            'orders': [{'quantity': quant, 'action': action, 'tickerId': optionId, 'tickerType': 'OPTION'}],
+            'orders': [{'quantity': int(quant), 'action': action, 'tickerId': int(optionId), 'tickerType': 'OPTION'}],
         }
 
-        if orderType == 'LMT' and lmtPrice:
+        if orderType == 'LMT' and lmtPrice :
             data['lmtPrice'] = float(lmtPrice)
-        if orderType == 'STP' and stpPrice:
+        elif orderType == 'STP' and stpPrice :
             data['auxPrice'] = float(stpPrice)
-        if orderType == 'STP LMT' and lmtPrice and stpPrice:
+        elif orderType == 'STP LMT' and lmtPrice and stpPrice :
             data['lmtPrice'] = float(lmtPrice)
             data['auxPrice'] = float(stpPrice)
 
         response = requests.post(self._urls.place_option_orders(self._account_id), json=data, headers=headers)
         if response.status_code != 200:
             raise Exception('place_option_order failed', response.status_code, response.reason)
-        return True
+        return response.text
 
-    def replace_option_order(self, order=None, lmtPrice=None, stpPrice=None, enforce=None, quant=0):
+    def modify_order_option(self, order=None, lmtPrice=None, stpPrice=None, enforce=None, quant=0):
         '''
         order: dict from get_current_orders
         stpPrice: float
         lmtPrice: float
-        enforce: DAY
+        enforce: GTC / DAY
         quant: int
         '''
         headers = self.build_req_headers(include_trade_token=True, include_time=True)
@@ -569,11 +727,11 @@ class webull:
                         'orderId': order['orderId']}]
         }
 
-        if order['orderType'] == 'LMT' and (lmtPrice or order['lmtPrice']):
+        if order['orderType'] == 'LMT' and (lmtPrice or order.get('lmtPrice')):
             data['lmtPrice'] = lmtPrice or order['lmtPrice']
-        if order['orderType'] and (stpPrice or order['auxPrice']):
+        elif order['orderType'] == 'STP' and (stpPrice or order.get('auxPrice')):
             data['auxPrice'] = stpPrice or order['auxPrice']
-        if order['orderType'] == 'STP LMT' and (stpPrice or order['auxPrice']) and (lmtPrice or order['lmtPrice']):
+        elif order['orderType'] == 'STP LMT' and (stpPrice or order.get('auxPrice')) and (lmtPrice or order.get('lmtPrice')):
             data['auxPrice'] = stpPrice or order['auxPrice']
             data['lmtPrice'] = lmtPrice or order['lmtPrice']
 
@@ -582,16 +740,22 @@ class webull:
             raise Exception('replace_option_order failed', response.status_code, response.reason)
         return True
 
+    def cancel_all_orders(self):
+        '''
+        Cancels all open (aka 'working') orders
+        '''
+        open_orders = self.get_current_orders()
+        for order in open_orders:
+            self.cancel_order(order['orderId'])
 
     def get_tradable(self, stock=''):
         '''
         get if stock is tradable
         '''
         headers = self.build_req_headers()
-        response = requests.get(self._urls.is_tradable(self.get_ticker(stock), headers=headers))
+        response = requests.get(self._urls.is_tradable(self.get_ticker(stock)), headers=headers)
         return response.json()
 
-
     def alerts_list(self):
         '''
         Get alerts
@@ -600,7 +764,10 @@ class webull:
 
         response = requests.get(self._urls.list_alerts(), headers=headers)
         result = response.json()
-        return result.get('data', [])
+        if 'data' in result:
+            return result.get('data', [])
+        else:
+            return None
 
     def alerts_remove(self, alert=None, priceAlert=True, smartAlert=True):
         '''
@@ -689,19 +856,19 @@ class webull:
             raise Exception('alerts_add failed', response.status_code, response.reason)
         return True
 
-    def get_active_gainer_loser(self, direction='gainer', count=20):
-        '''
-        gets active / gainer / loser stocks sorted by change
-        direction: active / gainer / loser
-        '''
-        headers = self.build_req_headers()
+    def active_gainer_loser(self, direction='gainer', rank_type='afterMarket', count=50) :
+          '''
+          gets gainer / loser / active stocks sorted by change
+          direction: gainer / loser / active
+          rank_type: preMarket / afterMarket / 5min / 1d / 5d / 1m / 3m / 52w (gainer/loser)
+                     volume / turnoverRatio / range (active)
+          '''
+          headers = self.build_req_headers()
 
-        params = {'regionId': 6, 'userRegionId': 6, 'pageSize': count}
-        response = requests.get(self._urls.active_gainers_losers(direction), params=params, headers=headers)
-        result = response.json()
-        result = sorted(result, key=lambda k: k['change'], reverse=True)
-        return result
+          response = requests.get(self._urls.active_gainers_losers(direction, self._region_code, rank_type, count), headers=headers)
+          result = response.json()
 
+          return result
 
     def run_screener(self, region=None, price_lte=None, price_gte=None, pct_chg_gte=None, pct_chg_lte=None, sort=None,
                      sort_dir=None, vol_lte=None, vol_gte=None):
@@ -748,6 +915,73 @@ class webull:
         headers = self.build_req_headers()
         return requests.get(self._urls.analysis(self.get_ticker(stock)), headers=headers).json()
 
+    def get_capital_flow(self, stock=None, tId=None, show_hist=True):
+        '''
+        get capital flow
+        :param stock:
+        :param tId:
+        :param show_hist:
+        :return: list of capital flow
+        '''
+        headers = self.build_req_headers()
+        if not tId is None:
+            pass
+        elif not stock is None:
+            tId = self.get_ticker(stock)
+        else:
+            raise ValueError('Must provide a stock symbol or a stock id')
+        return requests.get(self._urls.analysis_capital_flow(tId, show_hist), headers=headers).json()
+
+    def get_etf_holding(self, stock=None, tId=None, has_num=0, count=50):
+        '''
+        get ETF holdings by stock
+        :param stock:
+        :param tId:
+        :param has_num:
+        :param count:
+        :return: list of ETF holdings
+        '''
+        headers = self.build_req_headers()
+        if not tId is None:
+            pass
+        elif not stock is None:
+            tId = self.get_ticker(stock)
+        else:
+            raise ValueError('Must provide a stock symbol or a stock id')
+        return requests.get(self._urls.analysis_etf_holding(tId, has_num, count), headers=headers).json()
+
+    def get_institutional_holding(self, stock=None, tId=None):
+        '''
+        get institutional holdings
+        :param stock:
+        :param tId:
+        :return: list of institutional holdings
+        '''
+        headers = self.build_req_headers()
+        if not tId is None:
+            pass
+        elif not stock is None:
+            tId = self.get_ticker(stock)
+        else:
+            raise ValueError('Must provide a stock symbol or a stock id')
+        return requests.get(self._urls.analysis_institutional_holding(tId), headers=headers).json()
+
+    def get_short_interest(self, stock=None, tId=None):
+        '''
+        get short interest
+        :param stock:
+        :param tId:
+        :return: list of short interest
+        '''
+        headers = self.build_req_headers()
+        if not tId is None:
+            pass
+        elif not stock is None:
+            tId = self.get_ticker(stock)
+        else:
+            raise ValueError('Must provide a stock symbol or a stock id')
+        return requests.get(self._urls.analysis_shortinterest(tId), headers=headers).json()
+
     def get_financials(self, stock=None):
         '''
         get financials info and returns a dict of financial info
@@ -867,7 +1101,7 @@ class webull:
         '''
         rank = []
         headers = self.build_req_headers()
-        params = {'regionId': 6, 'userRegionId': 6, 'platform': 'pc', 'limitCards': 'latestActivityPc'}
+        params = {'regionId': self._region_code, 'userRegionId': self._region_code, 'platform': 'pc', 'limitCards': 'latestActivityPc'}
         response = requests.get(self._urls.rankings(), params=params, headers=headers)
         result = response.json()[0].get('data')
         if extendTrading:
@@ -880,6 +1114,16 @@ class webull:
                     rank = data['data']
         return rank
 
+    def get_watchlists(self):
+        """
+        get user watchlists
+        :return:
+        """
+        headers = self.build_req_headers()
+        params = {'version': 0}
+        response = requests.get(self._urls.portfolio_lists(), params=params, headers=headers)
+        return response.json()['portfolioList']
+
 
 
 ''' Paper support '''
@@ -899,8 +1143,12 @@ class paper_webull(webull):
         headers = self.build_req_headers()
         response = requests.get(self._urls.paper_account_id(), headers=headers)
         result = response.json()
-        id = result[0]['id']
-        return id
+        if result is not None and result[0] is not None:
+            id = result[0]['id']
+            self._account_id = id
+            return id
+        else:
+            return None
 
     def get_current_orders(self):
         ''' Open paper trading orders '''
